/* This file has been prepared for Doxygen automatic documentation generation.*/
/*! \file *********************************************************************
 *
 * \brief
 *      Charge functions
 *
 *      Contains the functions for charging with constant current and voltage,
 *      and for deciding when to halt.
 *
 * \par Application note:
 *      AVR458: Charging Li-Ion Batteries with BC100 \n
 *      AVR463: Charging NiMH Batteries with BC100
 *
 * \par Documentation
 *      For comprehensive code documentation, supported compilers, compiler
 *      settings and supported devices see readme.html
 *
 * \author
 *      Atmel Corporation: http://www.atmel.com \n
 *      Support email: avr@atmel.com
 *
 *
 * $Name$
 * $Revision: 4254 $
 * $RCSfile$
 * $URL: http://revisor.norway.atmel.com/AppsAVR8/avr458_Charging_Li-Ion_Batteries_with_BC100/trunk/code/IAR/chargefunc.c $
 * $Date: 2008-07-25 12:32:45 +0200 (fr, 25 jul 2008) $\n
 ******************************************************************************/

#include <ioavr.h>

#include "enums.h"
#include "structs.h"

#include "ADC.h"
#include "battery.h"
#include "chargefunc.h"
#include "main.h"
#include "menu.h"
#include "PWM.h"
#include "statefunc.h"
#include "time.h"

#ifdef NIMH
#include "NIMHspecs.h"
#endif // NIMH

#ifdef LIION
#include "LIIONspecs.h"
#endif // LIION


//******************************************************************************
// Variables
//******************************************************************************
//! Struct that holds parameters for ConstantCurrent() and ConstantVoltage().
ChargeParameters_t ChargeParameters;

//! Struct that holds parameters for HaltNow().
HaltParameters_t HaltParameters;


//******************************************************************************
// Functions
//******************************************************************************
/*! \brief Charges battery with a constant current.
 *
 * This function applies a constant current (set in ChargeParameters.Current)
 * to the battery until HaltNow() returns TRUE, or a PWM error occurs and
 * \ref ABORT_IF_PWM_MIN or \ref ABORT_IF_PWM_MAX is defined.\n
 * The charge current can vary with +/- \ref BAT_CURRENT_HYST.\n
 * If the Master inhibits charging, timers are stopped and PWM output dropped.
 * Once the battery is no longer flagged for charge inhibit, timers are
 * started again and charging resumed.
 *
 * \retval ChargeParameters.NextState Next state once this stage is done.
 * If no errors occured, this will be whatever was set in Charge(). Otherwise,
 * HaltNow() will have set a new next state.
 */
unsigned char ConstantCurrent(void)
{
	unsigned char error = FALSE,
	              wasStopped = FALSE;

	do	{
      // Wait for ADC conversions to complete.
		ADC_Wait();

		// If Master has flagged for a charge inhibit, pause charging.
		// (This is to prevent damage during prolonged serial communication.)
		if (BattControl[BattActive].ChargeInhibit) {
			wasStopped = TRUE;
			Time_Stop();
			OCR1B = 0;
		} else {
			// Continue charging!
			if (wasStopped) {
				wasStopped = FALSE;

				// Timer variables are not reset by this.
				Time_Start();
			}

			// Adjust the charge current to within ChargeParameters.Current
			// +/- BAT_CURRENT_HYST.
			if ((ADCS.avgIBAT < 0) ||
			    (ADCS.avgIBAT < (ChargeParameters.Current - BAT_CURRENT_HYST))) {

				if(!PWM_IncrementDutyCycle()) {
#ifdef ABORT_IF_PWM_MAX
				// If the duty cycle cannot be incremented, flag error and
				// go to error state.
					SetErrorFlag(ERR_PWM_CONTROL);
					ChargeParameters.NextState = ST_ERROR;
					error = TRUE;
#endif
				}
			} else if ((ADCS.avgIBAT >= 0) &&
				 (ADCS.avgIBAT > (ChargeParameters.Current + BAT_CURRENT_HYST))) {

				if(!PWM_DecrementDutyCycle()) {
#ifdef ABORT_IF_PWM_MIN
					// If the duty cycle cannot be decremented, flag error and
					// go to error state.
					SetErrorFlag(ERR_PWM_CONTROL);
					ChargeParameters.NextState = ST_ERROR;
					error = TRUE;
#endif
				}
			}
		}
	} while (!HaltNow() && !error);

	// Return the next state to Charge(). If an error has occured, this will
	// point to some other state than the next state of charging.
	return(ChargeParameters.NextState);
}


/*! \brief Charges battery with a constant voltage
 *
 * This function applies a constant voltage (set in ChargeParameters.Voltage)
 * to the battery until HaltNow() returns TRUE, or a PWM error occurs and
 * \ref ABORT_IF_PWM_MIN or \ref ABORT_IF_PWM_MAX is defined.\n
 * The charge voltage can vary with +/- \ref BAT_VOLTAGE_HYST.\n
 * If the Master inhibits charging, timers are stopped and PWM output dropped.
 * Once the battery is no longer flagged for charge inhibit, timers are
 * started again and charging resumed.
 *
 * \retval ChargeParameters.NextState Next state once this stage is done.
 * If no errors occured, this will be whatever was set in Charge(). Otherwise,
 * HaltNow() will have set a new next state.
 */
unsigned char ConstantVoltage(void)
{
	unsigned char error = FALSE,
	              wasStopped = FALSE;

	do{

		// Wait for ADC conversions to complete.
		ADC_Wait();

		// If Master has flagged for a charge inhibit, pause charging.
		// (This is to prevent damage during prolonged serial communication.)
		if (BattControl[BattActive].ChargeInhibit) {
			wasStopped = TRUE;
			Time_Stop();
			OCR1B = 0;
		}

		else {
			// Continue charging!
			if (wasStopped) {
				wasStopped = FALSE;

				// Timer variables aren't reset by this.
				Time_Start();
			}

			// Adjust the charge voltage to within ChargeParameters.Voltage
			// +/- BAT_VOLTAGE_HYST.
			if (ADCS.VBAT < (ChargeParameters.Voltage - BAT_VOLTAGE_HYST)) {

				if(!PWM_IncrementDutyCycle()) {
#ifdef ABORT_IF_PWM_MAX
				// Flag PWM control error and go to error-state if the duty
				// cycle cannot be incremented.
					SetErrorFlag(ERR_PWM_CONTROL);
					ChargeParameters.NextState = ST_ERROR;
					error = TRUE;
#endif
				}
			} else if (ADCS.VBAT > (ChargeParameters.Voltage + BAT_VOLTAGE_HYST)) {

				if(!PWM_DecrementDutyCycle()) {
#ifdef ABORT_IF_PWM_MIN
					// Flag PWM control error and go to error-state if duty
					// cycle cannot be decremented.
					SetErrorFlag(ERR_PWM_CONTROL);
					ChargeParameters.NextState = ST_ERROR;
					error = TRUE;
#endif
				}
			}
		}

	} while (!HaltNow() && !error);

	// Return the next state to Charge(). If an error has occured, this will
	// point to some other state than the next state of charging.
	return(ChargeParameters.NextState);
}


/*! \brief Charges battery with maximum allowable voltage or current
 *
 * This function charges in compliance with the Japanese regulations for
 * battery chargers, which specify maximum charge currents and voltages for
 * four temperature ranges:
 *
 * ID_|_Temp. range____|_Maximum I_|_Maximum U______
 * T0 |  0 - 10 deg. C | IC / IC/2 | 4.10 V / 4.25 V
 * T1 | 10 - 45 deg. C |     IC    | 4.25 V
 * T2 | 45 - 50 deg. C |     IC    | 4.15 V
 * T3 | 50 - 60 deg. C |     IC    | 4.10 V
 *
 * The maximum current or voltage for each temperature range is applied to the
 * battery until HaltNow() returns TRUE, or a PWM error occurs and
 * \ref ABORT_IF_PWM_MIN or \ref ABORT_IF_PWM_MAX is defined.\n
 *
 * Note that IC will depend on the type of battery, and may be less than 1 C.
 * This implementation uses whatever maximum current is set by the Resistor ID
 * lookup, which would be either something from the RID lookup-table, or the
 * default set in \ref battery.h (if charging without RID is enabled).
 *
 * Also, this function does not stop charging if the temperature is out
 * of range, i.e., below 0 or above 60 degrees C. Instead, this is done by
 * \ref HaltNow() which relies on the limits (\ref BAT_TEMPERATURE_MAX and 
 * \ref BAT_TEMPERATURE_IN) defined in \ref LIIONspecs.h or \ref NIMHspecs.h.
 *
 * If the Master inhibits charging, timers are stopped and PWM output dropped.
 * Once the battery is no longer flagged for charge inhibit, timers are
 * started again and charging resumed.
 *
 * \retval ChargeParameters.NextState Next state once this stage is done.
 * If no errors occured, this will be whatever was set in Charge(). Otherwise,
 * HaltNow() will have set a new next state.
 *
 * \note The temperature ranges are defined in \ref chargefunc.h ,
 * and the maximum current/voltages are defined in \ref LIIONspecs.h .
 *
 * \note To prevent "borderline" temperatures from causing the charging
 * parameters to constantly change, a temperature hysteresis, defined in
 * \ref chargefunc.h , is used.
 */
unsigned char MaxVoltageAndCurrent (void)
{
	unsigned char error = FALSE,
		      wasStopped = FALSE,
		      tempRangeChange;

	signed char lastTempRange = -1; // Must be set to an "illegal" temperature
	// in the first iteration.

	do {
		// Wait for ADC conversions to complete, then find the temperature.
		ADC_Wait();
		NTCLookUp();

		// If Master has flagged for a charge inhibit, pause charging.
		// (This is to prevent damage during prolonged serial communication.)
		if (BattControl[BattActive].ChargeInhibit) {
			wasStopped = TRUE;
			Time_Stop();
			OCR1B = 0;
		}
		else {
			// Continue charging!
			if (wasStopped) {
				wasStopped = FALSE;

				// Timer variables are not reset by this.
				Time_Start();
			}

         // Temperature hysteresis, allow some drift around the transitional values.
         // (Compare the new temperature with the hysteresis boundaries of the 
         // previously set temperature range, and flag for a change if the new 
         // temperature deviates beyond the hysteresis.
         tempRangeChange = FALSE;  // Start by assuming no change in temp. range.
         
         // See if the temperature range has changed, flag if it has. Handle T0!
         switch (lastTempRange) {
            case T0 :
               if ((BattData.Temperature - T0) > TEMP_HYST) {
                  tempRangeChange = TRUE;
               }
               else {
                  // We need to do this check if we're staying in this range:
                  // If the battery voltage is below threshold #1 for T0, attempt to
                  // regulate the charge current to max 1.0C and voltage to max 4.10 V.
                  // If the battery voltage is above threshold #1 for T0, attempt to
                  // regulate the charge current to max 0.5C and voltage to max 4.25 V.
                  
                  // This will cause a kind of "oscillation" starting when battery
                  // voltage reaches 4.10 V, and ending once the battery voltage
                  // surpasses 4.10 V for currents below 0.5 C, after which the
                  // function will keep applying 4.25 V.
                  // It will not cause any of the regulation limits to be crossed.
                  if (ADCS.VBAT <= BAT_VOLTAGE_MAX_T0) {
                     ChargeParameters.Voltage = BAT_VOLTAGE_MAX_T0;
                     ChargeParameters.Current = BattData.MaxCurrent;
                  }
                  else {
                     ChargeParameters.Voltage = BAT_VOLTAGE_MAX;
                     ChargeParameters.Current = BattData.MaxCurrent/2;
                  }
               }
            break;


            case T1 :
               if ((BattData.Temperature - T1) > TEMP_HYST
                     || (T0 - BattData.Temperature) > TEMP_HYST) {
                  tempRangeChange = TRUE;
               }
            break;


            case T2 :
               if ((BattData.Temperature - T2) > TEMP_HYST
                     || (T1 - BattData.Temperature) >= TEMP_HYST) {
                  tempRangeChange = TRUE;
               }
            break;


            case T3 :
               if ((T2 - BattData.Temperature) > TEMP_HYST) {
                  tempRangeChange = TRUE;
               }
            break;


            // This should only happen in first iteration.
            default : tempRangeChange = TRUE;
         }
         
         // If a change in temperature range was flagged, update charge parameters.
         if (tempRangeChange) {
            // Find new temperature range, set maximum charge current and voltage.
         
            if (BattData.Temperature <= T0) { // [-inf,T0] degrees C
               if (ADCS.VBAT <= BAT_VOLTAGE_MAX_T0) {
                  ChargeParameters.Voltage = BAT_VOLTAGE_MAX_T0;
                  ChargeParameters.Current = BattData.MaxCurrent;
               }
               else {
                  ChargeParameters.Voltage = BAT_VOLTAGE_MAX;
                  ChargeParameters.Current = BattData.MaxCurrent/2;
               }
               lastTempRange = T0;
            }
         
            else {
               // The other temperature ranges all allow charging at maximum 1.0 C,
               // but have different maximum voltages.
               ChargeParameters.Current = BattData.MaxCurrent;
         
               if (BattData.Temperature <= T1) { // (T0,T1]
                  ChargeParameters.Voltage = BAT_VOLTAGE_MAX;
                  lastTempRange = T1;
               }
               else if (BattData.Temperature <= T2) { // (T1,T2]
                  ChargeParameters.Voltage = BAT_VOLTAGE_MAX_T2;
                  lastTempRange = T2;
               }
               else { // (T2,inf]
                  ChargeParameters.Voltage = BAT_VOLTAGE_MAX_T3;
                  lastTempRange = T3;
               }
            }
         }

			// If both charge voltage and current are below limits, the PWM output may
			// be increased.
			if (ADCS.VBAT < (ChargeParameters.Voltage - BAT_VOLTAGE_HYST) &&
			   ADCS.avgIBAT < (ChargeParameters.Current - BAT_CURRENT_HYST)) {

				if(!PWM_IncrementDutyCycle()) {
#ifdef ABORT_IF_PWM_MAX
					// If the duty cycle cannot be incremented, flag error and
					// go to error state.
					SetErrorFlag(ERR_PWM_CONTROL);
					ChargeParameters.NextState = ST_ERROR;
					error = TRUE;
#endif
				}
				// If either voltage or current is above limit + hysteresis, reduce output.
			} 
			if (ADCS.VBAT > (ChargeParameters.Voltage + BAT_VOLTAGE_HYST) ||
			    ADCS.avgIBAT > (ChargeParameters.Current + BAT_CURRENT_HYST)) {
				if(!PWM_DecrementDutyCycle()) {
#ifdef ABORT_IF_PWM_MIN
					// If the duty cycle cannot be decremented, flag error and
					// go to error state.
					SetErrorFlag(ERR_PWM_CONTROL);
					ChargeParameters.NextState = ST_ERROR;
					error = TRUE;
#endif
				}
			}
		}
	} while (!HaltNow() && !error);

	// Return the next state to Charge(). If an error has occured, this will
	// point to some other state than the next state of charging.
	return(ChargeParameters.NextState);
}


/*! \brief Determines when to halt charging.
 *
 * This function evaluates parameters depending on what has been flagged in
 * HaltParameters.HaltFlags, and returns TRUE or FALSE if the charging should
 * halt or not.\n
 * In addition, error flagging on timeout (battery exhaustion) can be set.\n
 *
 * The function also checks if the battery temperature is within limits,
 * if mains is OK, and if BatteryCheck() returns TRUE.
 * If an error is detected, the associated errorflag is set and
 * ChargeParameters.NextState is changed to an appropriate state.
 *
 * \retval TRUE Halt now.
 * \retval FALSE Don't halt now.
 *
 * \note See chargefunc.h for definitions of halt flags.
 * \note It is generally a bad idea not to halt on a timeout.
 * \note If HALT_ON_VOLTAGE_DROP is set, HaltParameters.VBATMax should be
 * reset in Charge() before calling a charging-function.
 *
 * \todo "Priorities" of standard error checks OK?
 */
unsigned char HaltNow(void)
{
	unsigned char i, halt = FALSE;

	// Wait for a full ADC-cycle to finish.
	ADC_Wait();
	NTCLookUp();

	// Evaluate ADC readings according to HaltFlags. Flag errors if selected.
	// If an error is flagged, ChargeParameters.NextState is set to ST_ERROR.
	// (Gets overridden if either mains is failing, or the battery changes.)
	for (i = 0x01; i != 0; i <<= 1) {
		if (HaltParameters.HaltFlags & i) {
			switch (i) {
			// Is VBAT less than the recorded maximum?
			case HALT_VOLTAGE_DROP:

				// Update VBATMax if VBAT is higher or charge is inhibited by master.
    // Evaluate for halt otherwise.
				if (ADCS.VBAT > HaltParameters.VBATMax ||
				    BattControl[BattActive].ChargeInhibit) {
					HaltParameters.VBATMax = ADCS.VBAT;
				} else if((HaltParameters.VBATMax - ADCS.VBAT) >=
				          HaltParameters.VoltageDrop) {
					halt = TRUE;
				}
			break;


			// Has VBAT reached the maximum limit?
			case HALT_VOLTAGE_MAX:

				if (ADCS.VBAT >= HaltParameters.VoltageMax) {
					halt = TRUE;
				}
				break;


			// Has IBAT reached the minimum limit?
			case HALT_CURRENT_MIN:

    // If the charging current has dropped below the minimum threshold and
    // charging is not inhibited by master, flag a halt.
				if (ADCS.avgIBAT <= HaltParameters.CurrentMin &&
				    !BattControl[BattActive].ChargeInhibit) {
					halt = TRUE;
				}
				break;


			// Is the temperature rising too fast?
			case HALT_TEMPERATURE_RISE:

				// If rawNTC has increased, the temperature has dropped.
				// We can store this value for now, and start the timer.
				// Otherwise, check if NTC has changed too fast.
				if (ADCS.rawNTC > HaltParameters.LastNTC) {
					HaltParameters.LastNTC = ADCS.rawNTC;
					Time_Set(TIMER_TEMP,0,30,0);

				// Is the increase in temperature greater than the set threshold?
				} else  if ((HaltParameters.LastNTC - ADCS.rawNTC) >=
				  (BattData.ADCSteps * HaltParameters.TemperatureRise)) {

					// If this happened within a timeframe of 30 seconds, the
					// temperature is rising faster than we want.
					// If not, update LastNTC and reset timer.
					if (Time_Left(TIMER_TEMP))  {
						halt = TRUE;
					} else {
						HaltParameters.LastNTC = ADCS.rawNTC;
						Time_Set(TIMER_TEMP,0,30,0);
					}
				}
			break;


			// Is there any time left?
			case HALT_TIME:

				if (!Time_Left(TIMER_CHG)) {
					halt = TRUE;

					// If exhaustion flagging is selected, stop the PWM, disable the
					// battery and flag it as exhausted. Make ST_ERROR next state.
					if (HaltParameters.HaltFlags & HALT_FLAG_EXHAUSTION) {
							PWM_Stop();
							BattControl[BattActive].Enabled = FALSE;
							BattData.Exhausted = TRUE;
							SetErrorFlag(ERR_BATTERY_EXHAUSTED);
							ChargeParameters.NextState = ST_ERROR;
					}
				}
			break;


			default:  // Shouldn't end up here, but is needed for MISRA compliance.
			break;
			}
		}
	}

	// Standard checks:

	// Battery too cold or hot?
	if ((BattData.Temperature < HaltParameters.TemperatureMin) ||
	    (BattData.Temperature > HaltParameters.TemperatureMax)) {

		PWM_Stop();
		SetErrorFlag(ERR_BATTERY_TEMPERATURE);
		ChargeParameters.NextState = ST_ERROR;
		halt = TRUE;
	}

	// Battery not OK?
	if (!BatteryCheck()) {
		PWM_Stop();
		ChargeParameters.NextState = ST_INIT;
		halt = TRUE;
	}

	// Is mains voltage OK?
	if (!ADCS.Mains) {
		PWM_Stop();
		ChargeParameters.NextState = ST_SLEEP;
		halt = TRUE;
	}

	return(halt);
}
